#lang racket


(define (st-interp e)
  (define prims '(+ - * =))
  (define (prim? op) (if (member op prims) #t #f))
  
  (define (ret v st)
    (match st
      ['halt v]
      [`(letk ,x ,e1 ,env1 ,st1)
       (t-interp e1 (hash-set env1 x v) st1)]
      [`(ar ,e1 ,env1 ,st1)
       (t-interp e1 env1 `(fn ,v ,st1))]
      [`(fn (clo (lambda (,x) ,e1) ,env1) ,st1)
       (t-interp e1 (hash-set env1 x v) st1)]
      [`(fn (kont ,st2) ,st1)
       (ret v st2)]))
  
  (define (t-interp e env st)
    (match e
      [`(let ([,x ,e0]) ,e1)
       (t-interp e0 env `(letk ,x ,e1 ,env ,st))]
      
      [`(,(? prim? op) ,(? symbol? xs) ...)
       (ret (apply (eval op (make-base-namespace))
                   (map ((curry hash-ref) env) xs))
            st)]
      [(? number? n)
       (ret n st)]
      
      [`(call/cc (lambda (,x) ,e0))
       (t-interp e0 (hash-set env x `(kont ,st)) st)]
      [`(lambda (,x) ,e0)
       (ret `(clo (lambda (,x) ,e0) ,env) st)]
      [(? symbol? x)
       (ret (hash-ref env x) st)]
      [`(,e0 ,e1)
       (t-interp e0 env `(ar ,e1 ,env ,st))]))
  
  (t-interp e (hash) 'halt))




; Tests

(st-interp '((lambda (u) (u u)) (lambda (z) z)))

(st-interp '((lambda (a) a)
             (call/cc (lambda (k)
                        ((lambda (b) b)
                         (k (lambda (c) c)))))))

(st-interp '(((lambda (a) (lambda (b) (+ a b))) 1) 2))


(st-interp '(let ([x 2])
              (let ([y 3])
                (let ([z (call/cc (lambda (k) (let ([w (k 1)])
                                                (+ w 2))))])
                  (+ x y z)))))

